/* jmixer.cpp - part of MUSE
 * (c)2000 Denis Roio aka jaromil
 * c++ engine for multiple mp3 realtime stream mixing
 * originally based on the mp3mixer core from Scott Manley aka szyzyg
 */

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h> // this is for the send in ended_channel()
#include <sys/wait.h>
#include <termios.h>
#include <sys/time.h>

#include "libsync/synchronous.h"

extern "C" {
#include "libshout/shout.h"
#include "lame/lame.h"
#include "mpeg.h"
}

/* anybody knows wich size is better to use for feeding chunks?  */
#define BUF_SIZE 2048 // was 1152 (mp3 default framesize)
#define MAX_SAMPLE 32766
#define CLIP 30000
#define CLIP_A 7650756 // ((MAX_SAMPLE-CLIP)*(MAX_SAMPLE-CLIP))
#define CLIP_B -27234 // (MAX_SAMPLE-2*CLIP)
#define MAX_CHANNELS 8
#define MAX_PLAYLIST_ENTRY 150
#define MAX_PATH_SIZE 150
#define QUALITY 1 /* best=1 worse=9 */

struct livein {
  short inbuf[BUF_SIZE];
  int rate;
  int channels;
  bool on;
};

struct channel {
  short inbuf[BUF_SIZE];
  char *name;
  float volume;
  float speed;
  int rate;
  int channels;
  bool on;
  bool end;
  int stream[2];
  pid_t pid;
};

class Stream_mixer: public LTI::Object {

private:

  short audio_buffer[BUF_SIZE]; /* interleaved buffer */
  int process_buffer[BUF_SIZE];
  short work1[BUF_SIZE*40], work2[BUF_SIZE];
  short inbuf[BUF_SIZE*40];

  int dsp;
  bool fullduplex;
  shout_conn_t ice_conn;
  lame_global_flags enc_flags;

  /* they're still hardcoded and initialized into the constructor */  
  int sample_rate;
  int stereo;

  struct channel stream_channel[MAX_CHANNELS];
  struct livein live_channel;
  
  bool write_toicecast(short *sendbuf, long len);
  void close_soundcard();

  unsigned int get_audio(int ch);
  unsigned int get_livein();
  int spawn_stream(int ch);
  void ended_channel(int ch);

  void make_mono(short *stereo, short *mono, unsigned int num);
  void make_stereo(short *mono, short *stereo, unsigned int num);
  void sep_stereo(short *stereo, short *left, short *right, unsigned int num);
  void mix_stereo(short *left, short *right, short *stereo, unsigned int num);

  void int_compress_samples(short *in, short *out, unsigned int num, unsigned int grain);
  void int_stretch_samples(short *in, short *out, unsigned int num, unsigned int grain);

  void compress_samples(short *in, short *out,unsigned int num1,unsigned int num2);
  void stretch_samples(short *in, short *out,unsigned int num1,unsigned int num2);
  void change_time(short *in, short *out, unsigned int num_in, unsigned int num_out);
  void convert_audio(short *in,short *out,int ch1,int ch2,int n1,int n2);
  
public:
  Stream_mixer();

  void run();
  
  bool open_soundcard();
  bool open_icecast(char *ip, int port, char *pass, char *mountpoint, int bps);
  void play_chunks();
  void start_channel(int ch);
  void pause_channel(int ch);
  void stop_channel(int ch);
  void set_channel(int ch, int playlist_pos);
  void set_volume(int ch, float vol);
  void set_speed(int ch, float speed);
  void dump_status();
  void set_live(bool stat);

  /* this is for the libsync - registering synchronized method's ids */
  static const int id_set_channel = 1;
  static const int id_stop_channel = 2;
  static const int id_pause_channel = 3;
  static const int id_start_channel = 4;
  static const int id_set_volume = 5;
  static const int id_set_speed = 6;
  static const int id_dump_status = 7;
  static const int id_set_live = 8;

  bool dspout;
  bool icecastout;

  char playlist[MAX_PLAYLIST_ENTRY][MAX_PATH_SIZE];
  void set_playlist(char *pl);

  int *client;

  ~Stream_mixer();
};

Stream_mixer::Stream_mixer() {
  int c = 0;

  sample_rate=44100;
  stereo=1;

  fprintf(stderr,"Initializing channels ...");
  
  /* init channels */
  for(c=0;c<MAX_CHANNELS;c++) {
    stream_channel[c].volume=1.0;
    stream_channel[c].speed=1.0;
    stream_channel[c].rate=0;
    stream_channel[c].channels=0;
    stream_channel[c].on=false;
    stream_channel[c].end=false;
    //    stream_channel[c].stayon=1;
    stream_channel[c].stream[0]=0;
    stream_channel[c].stream[1]=0;
    stream_channel[c].pid=0;
    //    c_number++;
  }

  /* setup livein */
  live_channel.rate = sample_rate;
  live_channel.channels = 2;
  live_channel.on = false;

  fprintf(stderr," done.\n");

  /* Initialization as a synchronous object */
  synchronous.registerObjectName("Stream_mixer");

}

void Stream_mixer::run() {

  if(debug)
    cerr << selfName() << " running on thread " << (int)pthread_self() << "\n";

  while(1) {
    
    synchronous.select.accept(0, id_start_channel);
    synchronous.select.accept(1, id_stop_channel);
    synchronous.select.accept(2, id_set_channel);
    synchronous.select.accept(3, id_pause_channel);    
    synchronous.select.accept(4, id_set_volume);
    synchronous.select.accept(5, id_set_speed);
    synchronous.select.accept(6, id_dump_status);
    synchronous.select.accept(7, id_set_live);

    synchronous.select.waitUntil(8, LTI::Kernel::now()+5);
    synchronous.wait();

    play_chunks();
  }

  cerr << selfName() << " exiting " << "\n";

}

Stream_mixer::~Stream_mixer() {
  int i;

  for(i=0;i<MAX_CHANNELS;i++)
    stop_channel(i);

  if(dsp!=0)
    close_soundcard();
  
  if(icecastout)
    shout_disconnect(&ice_conn);

}

bool Stream_mixer::open_soundcard()
{
    int format,tstereo,speed,caps;

    if((dsp=open("/dev/dsp",O_RDWR))==-1) {
        perror("Error opening /dev/dsp: ");
	return(false);
    } else
      fprintf(stderr,"Found soundcard on /dev/dsp, initializing ...\n");

    format = AFMT_S16_LE;
    tstereo = stereo;
    speed = sample_rate;

    ioctl(dsp, SNDCTL_DSP_GETCAPS, &caps);
    if(caps & DSP_CAP_DUPLEX) {
      fullduplex = true;
      fprintf(stderr,"full duplex is supported, good\n");
      ioctl(dsp, SNDCTL_DSP_SETDUPLEX, 0);
      ioctl(dsp, DSP_CAP_DUPLEX, 0);
    } else {
      fprintf(stderr,"full duplex is not supported\n");
      fullduplex = false;
    }

    if(ioctl(dsp,SNDCTL_DSP_SETFMT,&format)==-1)
      fprintf(stderr,"failed to set data format\n");
    if(ioctl(dsp,SNDCTL_DSP_STEREO,&tstereo)==-1)
      fprintf(stderr,"something went wrong with the stereo setting\n");
    if(ioctl(dsp,SNDCTL_DSP_SPEED,&speed)==-1)
      fprintf(stderr,"speed setting failed\n");

    fprintf(stderr,"format setted to 16bit %dHz stereo.\n",sample_rate);
    dspout = true;

    return(true);
} /* open_soundcard */

bool Stream_mixer::open_icecast(char *ip, int port, char *pass, char *mountpoint, int bps) {
  shout_init_connection(&ice_conn);
  ice_conn.ip = ip;
  ice_conn.port = port;
  ice_conn.password = pass;
  ice_conn.mount = mountpoint;
  ice_conn.bitrate = bps*1000;
  
  fprintf(stderr,"Connecting to icecast on %s:%u ... ",ip,port);

  if(shout_connect(&ice_conn)) {
    fprintf(stderr,"estabilished mountpoint on /%s\n",mountpoint);
    /* initialize the lame mp3 encoder engine */
    lame_init(&enc_flags);
    if(stereo) enc_flags.num_channels = 2;
    enc_flags.num_channels = 2;
    enc_flags.in_samplerate = sample_rate;
    enc_flags.brate = bps;
    enc_flags.mode = 1; /* jstereo */
    enc_flags.quality = QUALITY;
    lame_init_params(&enc_flags);
    fprintf(stderr,"Mp3 encoder engine initialized.\n");

    icecastout = true;
    return(true);
  } else {
    fprintf(stderr,"ops!\nunable to connect or reach host (sorry, no details)\n");
    return(false);
  }
}

void Stream_mixer::play_chunks()
{
  int i,j,num;
  //  char temp[(BUF_SIZE<<4)/44100 + ((4*1152*16)/44100+512)];
  //  char temp[BUF_SIZE];
  num=BUF_SIZE;
  
  memset(process_buffer,0,num<<2);
  
  for(i=0;i<MAX_CHANNELS;i++) {
    if((stream_channel[i].pid)&&(stream_channel[i].on)) {
      get_audio(i);
      if(stream_channel[i].volume!=1.0)
	for(j=0;j<num;j++)
	  process_buffer[j] = (int) (process_buffer[j] + (stream_channel[i].inbuf[j] * stream_channel[i].volume));
      else
	for(j=0;j<num;j++)
	  process_buffer[j] = (int) (process_buffer[j] + stream_channel[i].inbuf[j]);

      /* if the mp3 reached the end */
      if(stream_channel[i].end)
	ended_channel(i);

    } // if (stream_channel[i].pid&&stream_channel[i].on)
  } // for i=0 to MAX_CHANNELS

  if(live_channel.on) {
    i = get_livein();
    for(j=0;j<num;j++)
      process_buffer[j] = (int) (process_buffer[j] + live_channel.inbuf[j]);
  }
  
  /* CLIPPING */
  for(j=0;j<num;j++)
    if((process_buffer[j]<CLIP)&&(process_buffer[j]>-CLIP))
      audio_buffer[j]=(short)(process_buffer[j]);
    else if (process_buffer[j]>0)
      audio_buffer[j]=MAX_SAMPLE-(CLIP_A/(CLIP_B+process_buffer[j]));
    else
      audio_buffer[j]=-(MAX_SAMPLE-(CLIP_A/(CLIP_B-process_buffer[j])));
  /* -------- */

  /* NO CLIPPING
     for(j=0;j<num;j++)
     audio_buffer[j]=(short)process_buffer[j];
    ----------- */

  if(dspout)
    write(dsp,audio_buffer,num<<1);

  if(icecastout) {
    num = lame_encode_buffer_interleaved(&enc_flags,&audio_buffer[0],num>>1,(char *)process_buffer,BUF_SIZE);
    if(num>0)
      write_toicecast((short int *)process_buffer,num);
    else if(num==-1) {
      fprintf(stderr,"Encoding failed: mp3buffer overflowed\n");
      exit(1);
    } 
  }
}

void Stream_mixer::start_channel(int ch) {
  stream_channel[ch].on = true;
  synchronous.reply();
}

void Stream_mixer::pause_channel(int ch) {
  stream_channel[ch].on = false;
  synchronous.reply();
}

void Stream_mixer::set_channel(int ch, int playlist_pos) {
  if(stream_channel[ch].pid) stop_channel(ch);
  stream_channel[ch].name = &playlist[playlist_pos][0];
  spawn_stream(ch);
  stream_channel[ch].on=true;
  stream_channel[ch].end=false;

  fprintf(stderr,"executed set_channel(%u, %u) as requested\n", ch, playlist_pos);
  
  synchronous.reply();
}

void Stream_mixer::set_volume(int ch, float vol) {
  stream_channel[ch].volume = vol;

  fprintf(stderr,"executed set_volume(%u, %f) as requested\n", ch, vol);

  synchronous.reply();
}

void Stream_mixer::set_speed(int ch, float speed) {
  stream_channel[ch].speed = speed;
  synchronous.reply();
}

void Stream_mixer::set_playlist(char *pl) {
  FILE *plist;
  int c = 0;
  char *p;

  plist = fopen(pl,"r");
  if(plist == NULL) {
    fprintf(stderr,"Playlist file not found\n");
    fprintf(stderr,"try: find /my/mp3/directory -name *.mp3 -print >playlist\n");
    exit(1);
  }

  while (fgets(&playlist[c][0],MAX_PATH_SIZE,plist) != NULL) {
    p = &playlist[c][0];
    while(*p!='\n') p++;
    *p = '\0';
    c++;
  }

  fprintf(stderr,"Got %u entries into the playlist, ready to rock!\n",c);

  fclose(plist);
}

void Stream_mixer::stop_channel(int ch) {
  if( stream_channel[ch].pid > 0) {
    if(!stream_channel[ch].end)
      kill(stream_channel[ch].pid,3); /* sends SIG QUIT */
    waitpid(stream_channel[ch].pid,NULL,0x0);

    close(stream_channel[ch].stream[0]);
    close(stream_channel[ch].stream[1]);

    stream_channel[ch].on=false;
    stream_channel[ch].pid=0;
    stream_channel[ch].stream[0] = 0;
    stream_channel[ch].stream[1] = 0;
  } else
    stream_channel[ch].on=false;
  
  /* it would be better a memset here */
  //    for(i=0;i<BUF_SIZE;i++) stream_channel[ch].inbuf[i]=0;
  memset(stream_channel[ch].inbuf,0,BUF_SIZE);

  synchronous.reply();
} /* stop_channel */

void Stream_mixer::dump_status() {
  int c;
  fprintf(stderr,"__________dumping_status__________\n");
  for(c=0;c<MAX_CHANNELS;c++) {
    
    if(!stream_channel[c].on)
      fprintf(stderr,"[channel %02u OFF] : PID %u\n",c,stream_channel[c].pid);
    else {
      fprintf(stderr,"[channel %02u ON ] : PID %u\n",c,stream_channel[c].pid);
      fprintf(stderr,"\tname:\t%s\n",stream_channel[c].name);
      fprintf(stderr,"\tvolume:\t%f\n",stream_channel[c].volume);
      fprintf(stderr,"\tspeed:\t%f\n",stream_channel[c].speed);
      fprintf(stderr,"\trate:\t%u\n",stream_channel[c].rate);
      fprintf(stderr,"\tchannels:\t%u\n",stream_channel[c].channels);
      fprintf(stderr,"\tend:\t%u\n",stream_channel[c].end);
      fprintf(stderr,"\tP stream:\t%u\n",stream_channel[c].stream[0]);
      fprintf(stderr,"\tC stream:\t%u\n",stream_channel[c].stream[1]);
    }
  }
  fprintf(stderr,"__________________________________\n");    
  synchronous.reply();
}

void Stream_mixer::set_live(bool stat) {
  if(!((dspout)&&(!fullduplex)&&(stat)))
    live_channel.on = stat;
  synchronous.reply();
}

unsigned int Stream_mixer::get_audio(int ch) {
  int i;
  unsigned int num_samples = 0;
  int opt;

  //  if(stream_channel[ch].on) {

    if(stereo)
      num_samples=(int)(((stream_channel[ch].rate*BUF_SIZE)*stream_channel[ch].speed)/(sample_rate<<1));
    else
      num_samples=(int)(((stream_channel[ch].rate*BUF_SIZE)*stream_channel[ch].speed)/sample_rate)
;

    opt = num_samples*stream_channel[ch].channels;

    i=read(stream_channel[ch].stream[0],inbuf,opt<<1);

    /* if we reach the end of the stream then ...
       Ok. It's a bug. it will be listed into the known bugs
       and it will be hopefully fixed. */
    if(i<opt<<1) {
      //      for(;i<opt;i++) inbuf[i]=0;
      //      fprintf(stderr,"UACK! detected end on channel %u receiving %u instead of %u\n",ch,i,opt);
      stream_channel[ch].end=true;
    }
    
    /* now convert the audio into something useful */
    if(stereo)
      convert_audio(inbuf,stream_channel[ch].inbuf,stream_channel[ch].channels,2,num_samples,BUF_SIZE>>1);
    else
      convert_audio(inbuf,stream_channel[ch].inbuf,stream_channel[ch].channels,1,num_samples,BUF_SIZE);
    //  } /* if(stream_channel[ch].on) */

  return num_samples;
} /* get_audio */

unsigned int Stream_mixer::get_livein() {
  unsigned int num_samples = 0;
  int opt, i;
  
  num_samples=(int)((live_channel.rate*BUF_SIZE)/(sample_rate<<1));
  opt = (num_samples*live_channel.channels)<<1;

  i=read(dsp,inbuf,opt);

  /* convert the audio into something useful */
  convert_audio(inbuf,live_channel.inbuf,live_channel.channels,2,num_samples,BUF_SIZE>>1);

  return(num_samples);
} /* get_livein */

int Stream_mixer::spawn_stream(int ch)
{
  int strm[2],trate,tchannels;
  pid_t pid;
  
  trate=format_of(stream_channel[ch].name,&tchannels);
  stream_channel[ch].rate=trate;
  stream_channel[ch].channels=tchannels;
  
  /* create pipe */
  pipe(strm);
  pid=fork();
  if(pid)
    {
      fprintf(stderr,"spawning mpg123 - got pid %u and pipe PAR:%u CHL:%u\n",pid,strm[0],strm[1]);
      stream_channel[ch].pid=pid;
      //      close(strm[1]);
      stream_channel[ch].stream[0]=strm[0];
      stream_channel[ch].stream[1]=strm[1];
      return(0);
    }
  else
    {
      close(strm[0]);
      close(1);
      dup(strm[1]);
      //      close(strm[0]);
      close(strm[1]);
      /* close audio too
	 don't use the close_soundcard because it causes glitches */
      close(dsp);
      execlp("mpg123","mpg123","-sq"/*,"-n 300"*/,stream_channel[ch].name,NULL);
    }
  return -1;
} /* spawn_stream */

void Stream_mixer::ended_channel(int ch) {
  /*  char temp[8];

      let it know to the guys out there
      if(*client>0) {
      sprintf(temp,"203\t%u\n\n",ch);
      send(*client, temp, strlen(temp+1), 0x0);
      } */

  close(stream_channel[ch].stream[0]);
  close(stream_channel[ch].stream[1]);
	
  stream_channel[ch].on=false;
  stream_channel[ch].end=false;
  stream_channel[ch].pid=0;
  stream_channel[ch].stream[0] = 0;
  stream_channel[ch].stream[1] = 0;

}  

bool Stream_mixer::write_toicecast(short *sendbuf, long len) {
  int res;
  res = shout_send_data(&ice_conn, (unsigned char *) sendbuf,len);
  if(!res)
    fprintf(stderr,"Send error: %i\n",ice_conn.error);
  shout_sleep(&ice_conn);
  return(res);
}

void Stream_mixer::close_soundcard() {
  ioctl(dsp, SNDCTL_DSP_RESET, 0);
  close(dsp);
}

void Stream_mixer::make_mono(short *stereo, short *mono, unsigned int num)
{
    unsigned int i;
    for(i=0;i<num;i++)
        mono[i]=(short)(((int)stereo[i<<1]+(int)stereo[(i<<1)+1])>>1);
} /* make_mono */

void Stream_mixer::make_stereo(short *mono, short *stereo, unsigned int num)
{
    unsigned int i;
    for(i=0;i<num;i++)
        stereo[1<<1]=stereo[(i<<1)+1]=mono[i];
}/* make_stereo */

void Stream_mixer::sep_stereo(short *stereo, short *left, short *right, unsigned int num)
{
    unsigned int i;
    for(i=0;i<num;i++)
    {
       	left[i]=stereo[i<<1];
       	right[i]=stereo[(i<<1)+1];
    }
} /* sep_stereo  */

void Stream_mixer::mix_stereo(short *left, short *right, short *stereo, unsigned int num)
{
    unsigned int i;
    for(i=0;i<num;i++)
    {
        stereo[i<<1]=left[i];
	stereo[(i<<1)+1]=right[i];
    }
} /* mix_stereo */

void Stream_mixer::int_compress_samples(short *in, short *out,unsigned int num,unsigned int grain)
{
    unsigned int i,j;
    int smpl;
	
    for(i=0;i<num;i++)
    {
       	smpl=in[i*grain];
	for(j=1;j<grain;j++) smpl+=in[i*grain+j];
	out[i]=(short)(smpl/grain);

    }
} /* int_compress_samples */

void Stream_mixer::int_stretch_samples(short *in,short *out,unsigned int num,unsigned int grain)
{
    unsigned int i,j;

    for(i=0;i<num;i++)
       	for(j=0;j<grain;j++)
            out[i*grain+j]=in[i];
	
} /* int_stretch_samples  */

void Stream_mixer::compress_samples(short *in, short *out,unsigned int num1,unsigned int num2)
{
    float step=(float)(num1)/num2;
    unsigned int i;

    for(i=0;i<num2;i++) out[i]=in[(int)(i*step)];
} /* compress_samples */

void Stream_mixer::stretch_samples(short *in, short *out,unsigned int num1,unsigned int num2)
{
    float step=(float)(num1)/num2;
    unsigned int i;

    for(i=0;i<num2;i++) out[i]=in[(int)(i*step)];
}/* stretch_samples */

/* changes the length of a buffer of samples uses fast shortcuts where possible */
void Stream_mixer::change_time(short *in, short *out, unsigned int num_in, unsigned int num_out){
    unsigned int i;
	
    if(num_in>num_out)
    {
        /* compress samples */
        if(num_in==(num_out<<1))
        {
            int_compress_samples(in,out,num_out,2);
        }
        else if( num_in==(num_out<<2))
        {
            int_compress_samples(in,out,num_out,4);
        }
        else
        {
            compress_samples(in,out,num_in,num_out);
        }
    }
    else if(num_in<num_out)
    {
        /* stretch samples */
        if((num_in<<1)==num_out)
        {
            int_stretch_samples(in,out,num_in,2);
        }
        else if ((num_in<<2)==num_out)
        {
            int_stretch_samples(in,out,num_in,4);
        }
        else
        {
            stretch_samples(in,out,num_in,num_out);
        }
    }
    else
    {
        for(i=0;i<num_out;i++)
            out[i]=in[i];
    }
} /* change_time   */

void Stream_mixer::convert_audio(short *in,short *out,int ch1,int ch2,int n1,int n2)
{
    int i;

    if((ch1==ch2) &&(n1==n2))
    {
      int gh = ch1*n1;
      for(i=0;i<gh;i++)
	out[i]=in[i]; 
        return;
    }

    if(ch2==1){
        /* output is mono */
        if(ch1==2){
            /* input is stereo */
            make_mono(in,work1,n1);
            change_time(work1,out,n1,n2);
        }
        else
        {
            /* input is mono */
            change_time(in,out,n1,n2);
        }
    }
    else
    {
        /* output is stereo */
        if(ch1==2){
            /* input is stereo */
            sep_stereo(  in,    work1, work2, n1    );
            change_time( work1, in,    n1,    n2    );
            change_time( work2, work1, n1,    n2    );
            mix_stereo(  in,    work1, out,   n2    );
        }
        else
        {
            /* input is mono */
            change_time(in,work1,n1,n2);
            make_stereo(work1,out,  n2);
        }
    }
} /* convert_audio */
